package p2.tictactoe;
import java.io.IOException;

/**
 * Class to enforce rules of TicTacToe.
 * @author $Author: oscar $
 * @version $Id: TicTacToe.java,v 1.4 2005/02/21 16:00:00 oscar Exp $
 */
public class TicTacToe implements BoardGame {
	static final int X = 0;
	static final int O = 1;

	private char[][] _gameState;	
	private Player _winner = new Player(); // nobody
	private Player[] _player;
	private int _turn = X; // initial turn
	private int _squaresLeft = 9;
	// hm ... should define a constant?
	
	/**
	 * The state of the game is represented as 3x3
	 * array of chars marked ' ', 'X', or 'O'.
	 * We index the state using chess notation,
	 * i.e., column is 'a' through 'c' and row is
	 * '1' through '3'.
	 */
	public TicTacToe(Player playerX, Player playerO)
	{
		_gameState = new char[3][3];
		for (char col='a'; col <='c'; col++)
			for (char row='1'; row<='3'; row++)
				this.set(col,row,' ');
		_player = new Player[2];
		_player[X] = playerX;
		_player[O] = playerO;
	}
	
	/**
	 * set() and get() translate between chess coordinates
	 * and array indices.
	 * NB: Use package scope.
	 */
	void set(char col, char row, char mark) {
		assert inRange(col, row);
		_gameState[col-'a'][row-'1'] = mark;
	}
	
	char get(char col, char row) {
		assert inRange(col, row);
		return _gameState[col-'a'][row-'1'];
	}
	
	/**
	 * The game is not over as long as there is no winner
	 * and somebody can still make a move ...
	 */
	public boolean notOver() {
		return this.winner().isNobody()
		&& this.squaresLeft() > 0;
	}
	
	/**
	 * A plain ascii representation of the game,
	 * mainly for debugging purposes.
	 */
	public String toString() {
		StringBuffer rep = new StringBuffer();
		
		for (char row='3'; row>='1'; row--) {
			rep.append(row);
			rep.append("  ");
			for (char col='a'; col <='c'; col++) {
				rep.append(this.get(col,row));
				if (col<'c') {
					rep.append(" | ");
				}
			}
			rep.append('\n');
			if (row>'1') {
				rep.append("  ---+---+---\n");
			}
		}
		rep.append("   a   b   c\n");
		return(rep.toString());
	}
	
	/**
	 * Needed for getter and setter preconditions.
	 * Reports true if coordinates are valid.
	 */
	boolean inRange(char col, char row) {
		return (('a'<=col) && (col<='c')
				&& ('1'<=row) && (row<='3'));
	}
	
	/**
	 * Called by the current player during an update().
	 * The player attempts to put a mark at a location.
	 */
	public void move(char col, char row, char mark)
	{
		assert this.notOver();
		assert inRange(col, row);
		assert this.get(col, row) == ' ';
		this.set(col, row, mark);
		this._squaresLeft--;
		this.swapTurn();
		this.checkWinner();
		assert this.invariant();
	}
	
	/**
	 * Ask the current player to make a move.
	 */
	public void update() throws IOException {
		this._player[_turn].move(this);
	}

	public Player currentPlayer() {
		return _player[_turn];
	}

	public Player winner() {
		return this._winner;
	}
	
	public int squaresLeft() {
		return this._squaresLeft;
	}
	
	private void swapTurn() {
		_turn = (_turn == X) ? O : X;
	}
	
	/**
	 * Check for a winning row, column or diagonal.
	 * (This code smells!  How can we clean it up?)
	 */
	private void checkWinner()
	{
		char player;
		for (char row='3'; row>='1'; row--) {
			player = this.get('a',row);
			if (player == this.get('b',row)
					&& player == this.get('c',row)) {
				this.setWinner(player);
				return;
			}
		}
		for (char col='a'; col <='c'; col++) {
			player = this.get(col,'1');
			if (player == this.get(col,'2')
					&& player == this.get(col,'3')) {
				this.setWinner(player);
				return;
			}
		}
		player = this.get('b','2');
		if (player == this.get('a','1')
				&& player == this.get('c','3')) {
			this.setWinner(player);
			return;
		}
		if (player == this.get('a','3')
				&& player == this.get('c','1')) {
			this.setWinner(player);
			return;
		}
	}
	
	/**
	 * Look up which player is the winner, and set _winner
	 * accordingly. Hm. Maybe we should store Players
	 * instead of chars in our array!
	 */
	private void setWinner(char player) {
		if (player == ' ')
			return;
		if (player == _player[X].mark())
			_winner = _player[X];
		else
			_winner = _player[O];
	}
	
	/**
	 * These seem obvious, which is exactly why
	 * they should be checked.
	 */
	private boolean invariant() {
		return (_turn == X || _turn == O)
			&& (this.notOver() 
				|| this.winner() == _player[X]
				|| this.winner() == _player[O]
				|| this.winner().isNobody())
			&& (_squaresLeft < 9
				// else, initially:
				|| _turn == X && this.winner().isNobody());
	}
}
