package p2.tictactoe;

import java.util.Observable;

/**
 * AbstractBoardGame implements common methods to
 * TicTacToe and Gomoku.
 * @author $Author: oscar $
 * @version $Id$
 */
public abstract class AbstractBoardGame
	extends Observable implements BoardGame
{
	static final int X = 0;
	static final int O = 1;

	protected int rows_;
	protected int cols_;
	protected int winningScore_;

	protected Player[][] gameState_;	
	protected final Player nobody_ = new InactivePlayer();
	protected Player winner_ = nobody_;
	protected Player[] player_;
	protected int turn_ = X; // initial turn
	protected int squaresLeft_;

	/**
	 * The state of the game is represented as 3x3
	 * array of chars marked ' ', 'X', or 'O'.
	 * We index the state using chess notation,
	 * i.e., column is 'a' through 'c' and row is
	 * '1' through '3'.
	 */
	public AbstractBoardGame(Player playerX, Player playerO)
	{
		player_ = new Player[2];
		player_[X] = playerX;
		player_[O] = playerO;
		playerX.setGame(this);
		playerO.setGame(this);
		
		this.init();
		
		squaresLeft_ = this.squares();
		gameState_ = new Player[rows_][cols_];
		for (int col=0; col <cols_; col++)
			for (int row=0; row<rows_; row++)
				this.set(col,row,nobody_);
	}
	
	/**
	 * Subclasses should initialize _rows, _cols
	 * and _winningScore
	 */
	protected abstract void init();

	/**
	 * @return Returns the _cols.
	 */
	public int get_cols() {
		return cols_;
	}
	
	/**
	 * @return Returns the _rows.
	 */
	public int get_rows() {
		return rows_;
	}

	/**
	 * Modifies the state of the game.  This method is only called
	 * by the BoardGame itself when a valid move has taken place.
	 * Any Observers are notified that the state has changed.
	 *
	 * @param col the column
	 * @param row the row
	 * @param player the Player attempting the move
	 */
	protected void set(int col, int row, Player player)
	{
		assert inRange(col, row);
		gameState_[col][row] = player;
	}

	public Player get(int col, int row)
	{
		assert inRange(col, row);
		return gameState_[col][row];
	}
	
	/**
	 * The game is not over as long as there is no winner
	 * and somebody can still make a move ...
	 */
	public boolean notOver() {
		return this.winner().isNobody()
		&& this.squaresLeft() > 0;
	}
	
	/**
	 * A plain ascii representation of the game,
	 * mainly for debugging purposes.
	 */
	public String toString() {
		StringBuffer rep = new StringBuffer();
	
		for (int row=this.rows_-1; row>=0; row--) {
			rep.append(1 + row);
			if (row < 9)
				rep.append(' '); // extra space for single digit
			rep.append("  ");
			for (int col=0; col < this.cols_; col++) {
				rep.append(this.get(col,row).mark());
				if (col<this.cols_-1) {
					rep.append(" | ");
				}
			}
			rep.append('\n');
			if (row>0) {
				rep.append("   ---");
				for (int i=1; i<this.cols_; i++)
					rep.append("+---");
				rep.append("\n");
			}
		}
		rep.append(" ");
		for (int col=0; col < this.cols_; col++) {
			rep.append("   ");
			rep.append((char) ('a' + col));
		}
		rep.append("\n");
		return(rep.toString());
	}
	
	/**
	 * Needed for getter and setter preconditions.
	 * Reports true if coordinates are valid.
	 */
	public boolean inRange(int col, int row) {
		return ((0<=col) && (col<this.cols_)
			&& (0<=row) && (row<this.rows_));
	}

	/**
	 * Called by the current player.  Here we check if the move
	 * is valid.  If so the game state is updated and the
	 * Observers are notified by calling this.set().
	 */
	public void move(int col, int row, Player p)
	{
		assert this.notOver();
		assert p == currentPlayer();
		assert this.get(col, row).isNobody();
		this.set(col, row, p);
		this.squaresLeft_--;
		this.swapTurn();
		this.checkWinner(col,row);
		assert this.invariant();
		setChanged();
		notifyObservers(new Move(col, row, p));
	}

	/**
	 * Extract the column letter from a String coordinate
	 * and convert to an index.
	 * E.g., "b17" -> 1
	 * NB: package scope for testing
	 */
	int getCol(String coord)
	{
		assert coord != null;
		assert coord.length() > 0;
		char col = coord.charAt(0);
		assert (col>='a') && (col<='z');
		return col - 'a';
	}
	
	/**
	 * Extract the column digits from a String coordinate
	 * and convert to an index.
	 * E.g., "b17" -> 16
	 * NB: package scope for testing
	 */
	int getRow(String coord)
	{
		assert coord != null;
		assert coord.length() > 1;
		try {
			int row = Integer.parseInt(coord.substring(1));
			return row - 1;
		} catch (NumberFormatException err) {
			throw new AssertionError(err.getMessage());
		}
	}

	public Player currentPlayer() {
		return player_[turn_];
	}

	public Player winner() {
		return this.winner_;
	}
	
	public int squaresLeft() {
		return this.squaresLeft_;
	}
	
	protected void swapTurn() {
		turn_ = (turn_ == X) ? O : X;
	}
	
	/**
	 * New algorithm needed for larger boards.
	 * Create a "Runner" that starts at the current position,
	 * and runs back in forth in a given direction, returning
	 * the length of the run (i.e., number of consecutive pieces
	 * of the same Player).  If the number is big enough, the
	 * current Player wins.
	 */
	protected void checkWinner(int col, int row)
	{
		Player player = this.get(col,row);
		Runner runner = new Runner(this, col, row);

		// check vertically
		if (runner.run(0,1) >= this.winningScore_)
			{ this.setWinner(player); return; }
		// check horizontally
		if (runner.run(1,0) >= this.winningScore_)
			{ this.setWinner(player); return; }
		// check diagonally
		if (runner.run(1,1) >= this.winningScore_)
			{ this.setWinner(player); return; }
		// and check the other diagonal
		if (runner.run(1,-1) >= this.winningScore_)
			{ this.setWinner(player); return; }
	}
		
	protected void setWinner(Player player) {
		winner_ = player;
	}
	
	/**
	 * These seem obvious, which is exactly why
	 * they should be checked.
	 */
	protected boolean invariant() {
		return (turn_ == X || turn_ == O)
			&& (this.notOver() 
				|| this.winner() == player_[X]
				|| this.winner() == player_[O]
				|| this.winner().isNobody())
			&& (squaresLeft_ < this.squares()
				// else, initially:
				|| turn_ == X && this.winner().isNobody());
	}
	
	/**
	 * @return total number of squares
	 * Package scope for testing.
	 */
	int squares() {
		return rows_ * cols_;
	}
}
