package p2.tictactoe.gui;

import p2.tictactoe.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * This class implements the view on a TicTacToe game.
 * The Frame contains a grid of Place objects that
 * intercept mouse clicks and forward them to the game's
 * Players.  The Frame is also an Observer for the game.
 * When the game notifies the Frame of state change, it
 * updates the corresponding Place.  Below the grid there
 * is a Label component to display messages.
 *
 * @author $Author: oscar $
 * @version $Id: GameGUI.java,v 1.2 2005/02/26 17:23:23 oscar Exp $
 */
public class GameGUI extends Frame implements Observer {
	static private final String imageDir = "p2/tictactoe/gui/images/";

	private BoardGame game_;
	private Label label_;
	private Place[][] places_;
	private Image[] images_;
	
	static private final int MINSIZE = 50;

	public GameGUI(String title, Image[] images)
		throws HeadlessException
	{
		super(title);
		assert images.length == 3;
		images_ = images;
		game_ = makeGame();
		game_.addObserver(this);
		this.setSize(MINSIZE*game_.get_cols(),
				MINSIZE*(1+game_.get_rows()));
		// setLayout(new BorderLayout()); // by default
		add("North", makeControls());
		add("Center", makeGrid());
		label_ = new Label();
		add("South", label_);
		showFeedBack(game_.currentPlayer().mark() + " plays");
		this.addWindowListener(new WindowAdapter(){
			public void windowClosing(WindowEvent e) {
				GameGUI.this.dispose();
			}
		});
		this.setVisible(true);
	}

	public static void main(String[] args) {
		new GameGUI("TicTacToe", getImages());
	}

	/**
	 * Get the X and O images from the file system.
	 */
	static private Image[] getImages() {
		Image[] images = new Image[3];
		images[0] = getImage(imageDir + "cross.gif");
		images[1] = getImage(imageDir + "not.gif");
		images[2] = getImage(imageDir + "blank.gif");
		return images;
	}

	static private Image getImage(String imageFile){
		Image image = Toolkit.getDefaultToolkit().createImage(imageFile);
		assert image != null;
		return image;
	}
	
	public BoardGame getGame() { return game_; }
	
	/**
	 * Instantiate the model.
	 * Need a way to select TicTacToe or Gomoku.
	 */
	private BoardGame makeGame() {
		Player X = new GUIplayer('X');
		Player O = new GUIplayer('O');
		return new TicTacToe(X, O);
		// return new Gomoku(X, O);
	}
	
	private Component makeControls() {
		Button again = new Button("New game");
		again.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				showFeedBack("starting new game ...");
				newGame();
			}
		});
		return again;
	}

	/**
	 * Make the grid that serves as a view and controller for the game.
	 */
	private Component makeGrid() {
		int cols = game_.get_cols();
		int rows = game_.get_rows();

		Panel grid = new Panel();
		grid.setLayout(new GridLayout(rows, cols));

		places_ = new Place[cols][rows];
		for (int row=rows-1; row>=0; row--) {
			for (int col=0; col<cols; col++) {
				Place p = new Place(col, row, images_);
				p.addMouseListener(new PlaceListener(p, this));
				grid.add(p);
				places_[col][row] = p;
			}
		}
		return grid;
	}
	
	/**
	 * Start a new game by deleting the old one, instantiating
	 * a new one, and clearing the images of the Places.
	 */
	private void newGame() {
		game_ = makeGame();
		game_.addObserver(this);

		int cols = game_.get_cols();
		int rows = game_.get_rows();

		for (int col=cols-1; col>=0; col--) {
			for (int row=0; row<rows; row++) {
				places_[col][row].clearImage();
			}
		}
		showFeedBack(game_.currentPlayer().mark() + " plays");
	}

	/**
	 * Displays a feedback string in the associated Label component.
	 * @param msg the string to display
	 */
	public void showFeedBack(String msg) {
		label_.setText(msg);
	}

	/**
	 * Implements the java.util.Observer interface.
	 * Called by the BoardGame when its state changes.
	 * @param o an instance of a BoardGame
	 * @param arg an instance of Move
	 * @see Move
	 */
	public void update(Observable o, Object arg) {
		Move move = (Move) arg;
		showFeedBack("got an update: " + move);
		places_[move.col][move.row].setMove(move.player);
	}
}

